import sys
import platform
import os
from typing import Optional, Dict

from PySide6 import QtWidgets
from PySide6.QtWidgets import QWidget, QDialog, QLineEdit

from sohstationviewer.view.util.plot_func_names import plot_functions

from sohstationviewer.database.process_db import execute_db
from sohstationviewer.database.extract_data import (
    get_param_info, create_assign_string_for_db_query
)

from sohstationviewer.conf.dbSettings import modify_db_path


class EditSingleParamDialog(QDialog):
    """
    Dialog to add info for channel not in database or edit the existing channel
    """
    def __init__(self, parent: Optional[QWidget],
                 param: str):
        """
        :param parent: the parent widget
        :param chan_id: name of channel to be added/edited
        :param data_type: type of the data being processed
        """
        self.param = param
        # # To skip on_param_chkbox_changed() when param is changed by the
        # # program at the beginning
        # self.param_changed_by_signal = False
        # database info of the channel
        self.channel_info: Dict = {}
        # database info of the channel's parameter
        self.param_info: Dict = {}
        super(EditSingleParamDialog, self).__init__(parent)

        # parameter's plot type which decides the shape of the plot
        self.plot_type_cbo_box = QtWidgets.QComboBox(self)
        self.plot_type_cbo_box.addItems([""] + list(plot_functions.keys()))

        # value color in black mode
        self.value_colorb_widget = QLineEdit(self)
        self.value_colorb_widget.setPlaceholderText(
            "Click edit button to add value color string")
        self.value_colorb_widget.setToolTip("Priority from left to right")
        # value, color in white mode
        self.value_colorw_widget = QLineEdit(self)
        self.value_colorw_widget.setPlaceholderText(
            "Click edit button to add value color string")
        self.value_colorw_widget.setToolTip("Priority from left to right")

        # height of the plot
        self.height_spnbox = QtWidgets.QSpinBox()
        self.height_spnbox.setMinimum(0)
        self.height_spnbox.setMaximum(8)
        self.height_spnbox.setToolTip("Relative height of the plot")
        self.height_warning_label = QtWidgets.QLabel(
            "(Height setting will only be applied after RePlot is clicked.)")
        self.height_warning_label.setStyleSheet(
            "QLabel {color: red; font-size: 10; font-style: italic;}"
        )

        # button to save change to DB
        self.save_param_btn = QtWidgets.QPushButton(
            "SAVE PARAMETER", self)
        # button to close dialog without doing anything
        self.cancel_btn = QtWidgets.QPushButton('CANCEL', self)

        self.setup_ui()
        self.set_param_info()
        self.connect_signals()

    def setup_ui(self) -> None:
        self.setWindowTitle(f"Edit Parameter {self.param}")

        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)

        param_layout = QtWidgets.QGridLayout()
        main_layout.addLayout(param_layout)

        param_layout.addWidget(QtWidgets.QLabel('Plot Type'), 0, 0, 1, 1)
        param_layout.addWidget(self.plot_type_cbo_box, 0, 1, 1, 1)

        param_layout.addWidget(QtWidgets.QLabel(
            'Value Color (black)'), 1, 0, 1, 1)
        param_layout.addWidget(self.value_colorb_widget, 1, 1, 1, 1)

        param_layout.addWidget(QtWidgets.QLabel(
            'Value Color (white)'), 2, 0, 1, 1)
        param_layout.addWidget(self.value_colorw_widget, 2, 1, 1, 1)

        param_layout.addWidget(QtWidgets.QLabel('Height'), 3, 0, 1, 1)
        param_layout.addWidget(self.height_spnbox, 3, 1, 1, 1)

        param_layout.addWidget(self.height_warning_label, 4, 0, 1, 2)
        param_layout.addWidget(self.cancel_btn, 5, 0, 1, 1)
        param_layout.addWidget(self.save_param_btn, 5, 1, 1, 1)

    def connect_signals(self) -> None:
        self.plot_type_cbo_box.currentTextChanged.connect(self.set_plot_type)
        self.cancel_btn.clicked.connect(self.close)
        self.save_param_btn.clicked.connect(self.on_save_param)

    def set_param_info(self) -> None:
        """
        Fill up all info boxes
        """
        self.param_info = get_param_info(self.param)
        self.set_plot_type(self.param_info['plotType'])
        self.height_spnbox.setValue(self.param_info['height'])

    def set_plot_type(self, plot_type: str) -> None:
        """
        Add Plot Type, Value Color strings.
        If there is no Plot Type, no Value Color or Height because no plot.
        :param plot_type: name of Plot Type
        """
        if plot_type in ["", None]:
            self.plot_type_cbo_box.setCurrentText('')
            self.value_colorb_widget.setEnabled(False)
            self.value_colorb_widget.clear()
            self.value_colorw_widget.setEnabled(False)
            self.value_colorw_widget.clear()
            self.height_spnbox.setValue(0)
        else:
            self.plot_type_cbo_box.setCurrentText(plot_type)
            value_color_b = self.param_info['valueColorsB']
            value_color_w = self.param_info['valueColorsW']
            self.value_colorb_widget.setText(value_color_b)
            self.value_colorw_widget.setText(value_color_w)

    def on_save_param(self):
        """
        Save parameter info to Parameters table
        """
        plot_type = create_assign_string_for_db_query(
            'plotType', self.plot_type_cbo_box.currentText())
        value_colorb = create_assign_string_for_db_query(
            'valueColorsB', self.value_colorb_widget.text())
        value_colorw = create_assign_string_for_db_query(
            'valueColorsW', self.value_colorw_widget.text())
        height = f"height={self.height_spnbox.value()}"
        sql = (f"UPDATE Parameters SET {plot_type}, {value_colorb}, "
               f"{value_colorw}, {height} WHERE param='{self.param}'")
        execute_db(sql)
        self.close()


if __name__ == '__main__':
    modify_db_path()
    os_name, version, *_ = platform.platform().split('-')
    if os_name == 'macOS':
        os.environ['QT_MAC_WANTS_LAYER'] = '1'
    app = QtWidgets.QApplication(sys.argv)

    # test linesDots. Ex: param: Input power supply current
    test = EditSingleParamDialog(None, 'Input power supply current')

    # test MultiColorDotsLowerBound. Ex: param:Backup volt
    # test = EditSingleParamDialog(None, 'Backup Volt', 'RT130')

    # test MultiColorDotsUpperBound. Ex: param:GNSS status
    # test = EditSingleParamDialog(None, 'GNSS status')

    # test UpDownDots. Ex: param. Ex: param:Net Up/down
    # test = EditSingleParamDialog(None, 'Net Up/Down', 'RT130')

    # test TriColorLInes. Ex: param. Ex: param:Error/warning
    # test = EditSingleParamDialog(None, 'Error/Warning', 'RT130')
    test.exec()
    sys.exit(app.exec())
