"""
param_dialog.py
GUI to add/dit/remove params
NOTE: Cannot remove or change params that are already used for channels.
"""
from PySide6 import QtWidgets, QtCore
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QComboBox, QWidget

from sohstationviewer.conf.constants import ColorMode, ALL_COLOR_MODES
from sohstationviewer.view.util.plot_func_names import plot_functions
from sohstationviewer.view.db_config.db_config_dialog import (
    UiDBInfoDialog,
)

from sohstationviewer.database.process_db import execute_db

from sohstationviewer.conf.dbSettings import dbConf


class ParamDialog(UiDBInfoDialog):
    def __init__(self, parent: QWidget, color_mode: ColorMode) -> None:
        """
        :param parent: QMainWindow/QWidget - the parent widget
        :param color_mode: the initial color mode of the dialog
        """
        self.color_mode = color_mode

        self.require_valuecolors_plottypes = [
            p for p in plot_functions.keys()
            if 'ValueColors' in plot_functions[p]['description']]
        super().__init__(
            parent,
            ['No.', 'Param', 'Plot Type', 'ValueColors', 'Height    '],
            'param', 'parameters',
            resize_content_columns=[0, 3])

        value_colors_column = 'valueColors' + self.color_mode
        self.insert_sql_template = (f"INSERT INTO Parameters "
                                    f"(param, plotType, {value_colors_column},"
                                    f" height) VALUES (?, ?, ?, ?)")
        self.update_sql_template = (f"UPDATE Parameters SET param=?, "
                                    f"plotType=?, {value_colors_column}=?, "
                                    f"height=? "
                                    f"WHERE param='%s'")

        self.setWindowTitle("Edit/Add/Delete Parameters")
        self.add_color_selector(color_mode)

    def add_color_selector(self, initial_color_mode: ColorMode) -> None:
        """
        Add a color mode selector widget to the dialog.

        :param initial_color_mode: the color mode to use as the initial
        selection when the dialog is shown.
        """
        color_mode_label = QtWidgets.QLabel('Color mode:')
        color_selector = QComboBox()
        color_selector.insertItem(0, initial_color_mode)
        other_color_modes = set(ALL_COLOR_MODES.keys()) - {initial_color_mode}
        color_selector.insertItems(1, other_color_modes)
        color_selector.setFixedWidth(100)
        color_selector.currentTextChanged.connect(self.on_color_mode_changed)
        self.bottom_layout.addWidget(color_mode_label, 0, 1, 1, 1,
                                     Qt.AlignRight)
        self.bottom_layout.addWidget(color_selector, 0, 2, 1, 1)

    def set_row_widgets(self, row_idx: int, fk: bool = False) -> None:
        """
        Set the widgets in a row in self.data_table_widgets.

        :param row_idx: index of row to add
        :param fk: bool: True if there is a foreign constrain that prevents the
            row to be deleted
        """
        self.add_widget(None, row_idx, 0)  # No.
        self.add_widget(self.database_rows, row_idx, 1, foreign_key=fk)
        plot_type = self.add_widget(
            self.database_rows, row_idx, 2,
            choices=[''] + sorted(plot_functions.keys()))
        place_holder_text = ""
        if plot_type in self.require_valuecolors_plottypes:
            place_holder_text = "Enter ValueColors..."
            if plot_type == 'linesDots':
                place_holder_text = "Ex: L:G|D:W"
            elif plot_type == 'multiColorDots':
                place_holder_text = "Ex: -1:_|0:R|2.3:Y|+2.3:G"
            elif plot_type == "upDownDots":
                place_holder_text = "Ex: 1:R|0:Y"
            elif plot_type == "dotForTime":
                place_holder_text = "Ex: G"
        self.add_widget(self.database_rows, row_idx, 3,
                        place_holder_text=place_holder_text)
        self.add_widget(self.database_rows, row_idx, 4, range_values=[0, 10])
        self.add_delete_button_to_row(row_idx, fk)

    def get_data_list(self):
        """
        Get list of data to fill self.data_table_widgets' content
        """
        # The valueColors for each color mode is stored in a separate column.
        # Seeing as we only need one of these columns for a color mode, we only
        # pull the needed valueColors column from the database.
        value_colors_column = 'valueColors' + self.color_mode
        param_rows = execute_db(
            f"SELECT param, "
            f"IFNULL(plotType, '') AS plotType, "
            f"IFNULL({value_colors_column}, '') AS valueColors, "
            f"IFNULL(height, 0) AS height FROM Parameters")
        return [[d[0], d[1], d[2], int(d[3])]
                for d in param_rows]

    def get_row_inputs(self, row_idx):
        """
        Get content from a row of widgets.
        The valueColors string's format is checked here before entered into DB

        :param row_idx: index of row
        """
        return [
            self.data_table_widget.cellWidget(row_idx, 1).text().strip(),
            self.data_table_widget.cellWidget(row_idx, 2).currentText(),
            self.data_table_widget.cellWidget(row_idx, 3).text().strip(),
            int(self.data_table_widget.cellWidget(row_idx, 4).value())
        ]

    def validate_row(self, row_id, row_content):
        """
        Check if the given row is valid. Invalid rows are those whose
        value colors does not have the correct format for the plot type.

        :return:
            - Whether the given row is valid
            - A message of what the issue is; is the empty string if there is
                no issue
        """
        is_valid_general_row, msg = super().validate_row(row_id, row_content)
        if not is_valid_general_row:
            return is_valid_general_row, msg

        plot_type = row_content[1]
        value_colors_string = row_content[2]

        if value_colors_string == "" and plot_type == 'multiColorDots':
            err_msg = (f"Row {row_id}: multiColorDots type requires some "
                       f"ValueColors value.")
            return False, err_msg

        if value_colors_string != "":
            if plot_type in self.require_valuecolors_plottypes:
                value_colors = value_colors_string.split("|")
                for vc in value_colors:
                    vc = vc.strip()
                    plot_type_info = dbConf.get(plot_type)
                    if plot_type_info is None:
                        continue
                    if not plot_type_info['pattern'].match(vc):
                        msg = (f"Row {row_id}: The ValueColors in this row is "
                               f"'{vc}', which doesn't match the required "
                               f"format.\n\n"
                               f"{plot_type_info['instruction']}")
                        return False, msg
            else:
                if plot_type == "":
                    msg = (f"Row {row_id}: No ValueColors should be entered "
                           f"when no Plot Type is selected.")
                else:
                    msg = (f"Row {row_id}: No ValueColors is required for "
                           f"Plot Type {plot_type}.")
                return False, msg
        return True, ''

    @QtCore.Slot()
    def on_color_mode_changed(self, new_color_mode: ColorMode):
        """
        Slot called when the color mode is changed in the color mode selector.

        :param new_color_mode: the new color mode
        """
        old_value_colors_column = 'valueColors' + self.color_mode
        self.color_mode = new_color_mode
        new_value_colors_column = 'valueColors' + self.color_mode
        self.insert_sql_template = self.insert_sql_template.replace(
            old_value_colors_column,
            new_value_colors_column)
        self.update_sql_template = self.update_sql_template.replace(
            old_value_colors_column,
            new_value_colors_column)

        # Remove all rows in the table while keeping the widths of the columns
        # intact
        self.data_table_widget.setRowCount(0)
        # Repopulates the table with data from the database. A custom routine
        # could have been written to replace only the ValueColors column in the
        # table, but that would take too much time for only a marginal increase
        # in performance, especially considering the Parameters table is
        # expected to be pretty small.
        self.update_data_table_widget_items()
